## -----------------------------------------------------------------------------------------------------------
## Code corresponding to: Forecasting Generalized Quantiles of Electricity Demand: A Functional Data Approach
## Generates forecasts for the TSO Amprion using functional data analysis (FDA) of generalized quantiles
## Evaluates results of FDA and the deterministic seasonal component (DSC) 
## -----------------------------------------------------------------------------------------------------------

rm(list = ls())
graphics.off()
# setwd("...\\Data and Command Files") # Set working directory

source("Code\\Data TSO.R") # Data Set: TSO Amprion
source("Code\\Functions.R")

libraries = c("expectreg","fda.usc","sm","vars")
lapply(libraries,function(x)if(!(x %in% installed.packages())){install.packages(x)})
lapply(libraries,require,quietly=TRUE,character.only=TRUE)

## Set parameter
n           = 702 # insample size
h           = 351 # forecast horizon
tau         = c(0.01,0.05,0.25,0.5,0.75,0.95,0.99) # expectile levels
p           = 1   # steps ahead

Forecast_DSC = matrix(nrow = h, ncol = 96)
Forecast_FDA = array(NA,dim=c(h, 96, 7))

## Forecasting (this takes a while)

for(k in 0:(h-p)){
  training      = Load_mat[1:(n+k),-1]
  training_temp = Temperature[1:(n+k+p)] 
  training_sun  = Sun[1:(n+k+p)]
  Load_list     = (unlist(as.data.frame(t(training))))
  
  tempseq = as.numeric(rep(rep(c(1:7),each=1),length.out=n+k+p))
  WD      = sapply(1:6, FUN = function(num){as.numeric(tempseq==num)})
  DSC     = dsc(Load_list, WD, dummy, n = n, k = k, hours = 96, p)
  DT_Load = DSC[[1]]
  Forecast_DSC[k+p,] = t(DSC[[2]])

resid_data_yearly = matrix(DT_Load,nrow=96)
x   = seq(1/nrow(resid_data_yearly),1,length.out=nrow(resid_data_yearly)) # generate x
fit = array(0,dim=c(ncol(resid_data_yearly),96,7))

# Estimation of expectiles (this takes a while)
for(i in 1:ncol(resid_data_yearly)){
  (y = expectreg.ls(resid_data_yearly[,i]~rb(x,"pspline"),estimate="sheet",expectiles=tau,smooth="gcv"))
  fit[i,,] = y$fitted
}

result  = sapply(X = c(1: 7), FUN = fpca, fit = fit, x = x)
scores  = result[3,]

resid_temp = season.cov(training_temp, 351)
resid_sun  = season.cov(training_sun, 351)

exo           = cbind(resid_temp[1:(n+k)],resid_sun[1:(n+k)])
colnames(exo) = c("Temperature","Sunshine")
exo_fore      = cbind(resid_temp[(n+k+1):(n+k+p)],resid_sun[(n+k+1):(n+k+p)])

forecast.score  = lapply(1:7, FUN = var.forecast, scores = scores, exo = exo, exo_fore = exo_fore, p = p)
Forecast_FDA[k+p,,] = sapply(1:7, FUN = forecast.curve, result = result, forecast.score = forecast.score, season_forecast = Forecast_DSC, p = p)
}

## Evaluate Results

# Table 6: Evaluation FDA, DSC, TSO
Eval = lapply(list(Forecast_FDA[,,4], Forecast_DSC, Load_pred[,-1]), FUN = EVAL, Load = Load_mat[,-1], h = h, n = n, p = p)
FDA.mean.eval = Eval[[1]]
DSC.eval = Eval[[2]]
Amp.eval = Eval[[3]]

# Table 7: Evaluation FDA for tau
MSE_FDA = sapply(1:7, FUN = MSE.FDA, Fore = Forecast_FDA, Load = Load_mat, tau = tau, n = n, h = h, p = p)
rownames(MSE_FDA) = c("Mean", "SD")
colnames(MSE_FDA) = tau
MSE_FDA 